#    Copyright 2021, 2022, 2023 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from abc import ABCMeta, abstractmethod
from typing import (
    Iterable,
    List,
    Optional,
    Sequence,
    Type,
    TypeVar,
    Union,
)


from wsidicom.errors import WsiDicomMatchError
from wsidicom.geometry import SizeMm
from wsidicom.group import Group, Level
from wsidicom.instance import ImageType, WsiDataset, WsiInstance
from wsidicom.uid import SlideUids

SeriesType = TypeVar("SeriesType")


class Series(metaclass=ABCMeta):
    """Represents a series of Groups with the same image flavor, e.g.
    pyramidal levels, labels, or overviews.
    """

    def __init__(self, groups: Iterable[Group]):
        """Create a Series from list of Groups.

        Parameters
        ----------
        groups: Iterable[Group]
            List of groups to include in the series.
        """
        self._groups = list(groups)
        self._group_type = Group
        if len(self.groups) != 0 and self.groups[0].uids is not None:
            self._uids = self._validate_series(self.groups)
        else:
            self._uids = None

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self.groups})"

    def __str__(self) -> str:
        return f"{type(self).__name__} of groups {self.groups}"

    def __getitem__(self, index: int) -> Group:
        """Get group by index.

        Parameters
        ----------
        index: int
            Index in series to get

        Returns
        ----------
        Group
            The group at index in the series
        """
        return self.groups[index]

    def __len__(self) -> int:
        return len(self.groups)

    @property
    @abstractmethod
    def image_type(self) -> ImageType:
        """Should return the wsi type of the series ('VOLUME', 'LABEL', or
        'OVERVIEW'"""
        raise NotImplementedError()

    @property
    def groups(self) -> List[Group]:
        """Return contained groups."""
        return self._groups

    @property
    def uids(self) -> Optional[SlideUids]:
        """Return uids."""
        return self._uids

    @property
    def mpps(self) -> List[SizeMm]:
        """Return contained mpp (um/px)."""
        return [group.mpp for group in self.groups if group.mpp is not None]

    @property
    def datasets(self) -> List[WsiDataset]:
        """Return contained datasets."""

        series_datasets = [series.datasets for series in self.groups]
        return [dataset for sublist in series_datasets for dataset in sublist]

    @property
    def instances(self) -> List[WsiInstance]:
        """Return contained instances"""
        series_instances = [series.instances.values() for series in self.groups]
        return [instance for sublist in series_instances for instance in sublist]

    @classmethod
    def open(
        cls: Type[SeriesType], instances: Iterable[WsiInstance]
    ) -> Optional[SeriesType]:
        """Return series created from instances.

        Parameters
        ----------
        instances: Iterable[WsiInstance]
            Instances to create series from.

        Returns
        ----------
        Optional[SeriesType]
            Created series.
        """
        groups = Group.open(instances)
        if len(groups) == 0:
            return None
        return cls(groups)

    def _validate_series(
        self, groups: Union[Sequence[Group], Sequence[Level]]
    ) -> Optional[SlideUids]:
        """Check that no files or instances in series is duplicate and that
        all groups in series matches.
        Raises WsiDicomMatchError otherwise.
        Returns the matching base uid. If list of groups is empty, return None.

        Parameters
        ----------
        groups: Union[Sequence[Group], Sequence[Level]]
            List of groups or levels to check

        Returns
        ----------
        Optional[SlideUids]:
            Matching uids
        """
        WsiDataset.check_duplicate_dataset(self.datasets, self)
        WsiInstance.check_duplicate_instance(self.instances, self)

        try:
            base_group = groups[0]
            if base_group.image_type != self.image_type:
                raise WsiDicomMatchError(str(base_group), str(self))
            for group in groups[1:]:
                if not group.matches(base_group):
                    raise WsiDicomMatchError(str(group), str(self))
            return base_group.uids
        except IndexError:
            return None


class Overviews(Series):
    """Represents a series of Groups of the overview wsi flavor."""

    @property
    def image_type(self) -> ImageType:
        return ImageType.OVERVIEW


class Labels(Series):
    """Represents a series of Groups of the label wsi flavor."""

    @property
    def image_type(self) -> ImageType:
        return ImageType.LABEL



#    Copyright 2021, 2022, 2023 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from pathlib import Path
from typing import List, OrderedDict, Sequence, Union

from pydicom.uid import UID

from wsidicom.codec import Codec
from wsidicom.errors import WsiDicomNotFoundError
from wsidicom.file.wsidicom_file import WsiDicomFile
from wsidicom.instance import WsiDicomImageData


class WsiDicomFileImageData(WsiDicomImageData):
    """
    Represents image data read from dicom file(s).

    Image data can be sparsly or fully tiled and/or concatenated.
    """

    def __init__(self, files: Union[WsiDicomFile, Sequence[WsiDicomFile]]) -> None:
        """
        Create WsiDicomFileImageData from frame data in files.

        Parameters
        ----------
        files: Union[WsiDicomFile, Sequence[WsiDicomFile]]
            Single or list of WsiDicomFiles containing frame data.
        """
        if not isinstance(files, Sequence):
            files = [files]

        # Key is frame offset
        self._files = OrderedDict(
            (file.frame_offset, file)
            for file in sorted(files, key=lambda file: file.frame_offset)
        )
        self._transfer_syntax = files[0].transfer_syntax
        dataset = files[0].dataset
        codec = Codec.create(
            self.transfer_syntax,
            dataset.samples_per_pixel,
            dataset.bits,
            dataset.tile_size,
            dataset.photometric_interpretation,
        )
        super().__init__([file.dataset for file in self._files.values()], codec)

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self._files.values()})"

    def __str__(self) -> str:
        return f"{type(self).__name__} of files {self._files.values()}"

    @property
    def files(self) -> List[Path]:
        return [
            file.filepath for file in self._files.values() if file.filepath is not None
        ]

    @property
    def transfer_syntax(self) -> UID:
        """The uid of the transfer syntax of the image."""
        return self._transfer_syntax

    def _get_file(self, frame_index: int) -> WsiDicomFile:
        """
        Return file containing frame index.

        Raises WsiDicomNotFoundError if frame is not found.

        Parameters
        ----------
        frame_index: int
             Frame index to get

        Returns
        ----------
        WsiDicomFile
            File containing the frame
        """
        for frame_offset, file in self._files.items():
            if (
                frame_index < frame_offset + file.frame_count
                and frame_index >= frame_offset
            ):
                return file

        raise WsiDicomNotFoundError(f"Frame index {frame_index}", "instance")

    def _get_tile_frame(self, frame_index: int) -> bytes:
        """Return tile frame for frame index.

        Parameters
        ----------
        frame_index: int
             Frame index to get

        Returns
        ----------
        bytes
            The frame in bytes
        """
        file = self._get_file(frame_index)
        tile_frame = file.read_frame(frame_index)
        return tile_frame



#    Copyright 2021, 2022, 2023 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import threading
from functools import cached_property
from pathlib import Path
from struct import unpack
from typing import BinaryIO, List, Optional, Tuple, Union, cast

from pydicom.errors import InvalidDicomError
from pydicom.filebase import DicomFileLike
from pydicom.filereader import _read_file_meta_info, read_partial, read_preamble
from pydicom.tag import BaseTag, ItemTag, SequenceDelimiterTag, Tag
from pydicom.uid import UID, UncompressedTransferSyntaxes
from wsidicom.codec import Codec

from wsidicom.errors import WsiDicomFileError, WsiDicomNotSupportedError
from wsidicom.file.wsidicom_file_base import OffsetTableType, WsiDicomFileBase
from wsidicom.instance import ImageType, WsiDataset
from wsidicom.uid import FileUids


class WsiDicomFile(WsiDicomFileBase):
    """Represents a DICOM file (potentially) containing WSI image and metadata."""

    def __init__(
        self, stream: BinaryIO, filepath: Optional[Path] = None, owned: bool = False
    ):
        """
        Parse DICOM file in stream. If valid WSI type read required
        parameters. Parses frames in pixel data but does not read the frames.

        Parameters
        ----------
        stream: BinaryIO
            Stream to open.
        filepath: Optional[Path] = None
            Optional filepath of stream.
        owned: bool = False
            If the stream should be closed by this instance.
        """
        self._lock = threading.Lock()
        try:
            stream.seek(0)
            read_preamble(stream, False)
        except InvalidDicomError:
            raise WsiDicomFileError(stream, "is not a DICOM file.")
        file_meta = _read_file_meta_info(stream)
        self._transfer_syntax_uid = UID(file_meta.TransferSyntaxUID)
        super().__init__(stream, filepath, owned)
        self._file.is_little_endian = self._transfer_syntax_uid.is_little_endian
        self._file.is_implicit_VR = self._transfer_syntax_uid.is_implicit_VR
        extended_offset_table_tag = Tag("ExtendedOffsetTable")

        def _stop_at(tag: BaseTag, vr: Optional[str], length: int) -> bool:
            return tag >= extended_offset_table_tag

        self._file.seek(0)
        dataset = read_partial(
            cast(BinaryIO, self._file),
            _stop_at,
            defer_size=None,
            force=False,
            specific_tags=None,
        )
        self._pixel_data_position = self._file.tell()

        self._image_type = WsiDataset.is_supported_wsi_dicom(dataset)
        if self._image_type is not None:
            self._dataset = WsiDataset(dataset)
        else:
            raise WsiDicomNotSupportedError(f"Non-supported file {self._file}.")
        syntax_supported = Codec.is_supported(
            self.transfer_syntax,
            self._dataset.samples_per_pixel,
            self._dataset.bits,
            self._dataset.photometric_interpretation,
        )
        if not syntax_supported:
            raise WsiDicomNotSupportedError(
                f"Non-supported transfer syntax {self.transfer_syntax}"
            )

    def __str__(self) -> str:
        return self.pretty_str()

    @classmethod
    def open(cls, file: Path) -> "WsiDicomFile":
        """
        Open file in path as WsiDicomFile.
        """
        stream = open(file, "rb")
        return cls(stream, file, True)

    @cached_property
    def offset_table_type(self) -> OffsetTableType:
        """Return type of the offset table, or None if not present."""
        return self._get_offset_table_type()

    @property
    def dataset(self) -> WsiDataset:
        """Return pydicom dataset of file."""
        return self._dataset

    @property
    def image_type(self) -> Optional[ImageType]:
        return self._image_type

    @property
    def uids(self) -> FileUids:
        """Return uids"""
        return self.dataset.uids

    @property
    def transfer_syntax(self) -> UID:
        """Return transfer syntax uid"""
        return self._transfer_syntax_uid

    @property
    def frame_offset(self) -> int:
        """Return frame offset (for concatenated file, 0 otherwise)"""
        return self.dataset.frame_offset

    @cached_property
    def frame_positions(self) -> List[Tuple[int, int]]:
        """Return frame positions and lengths"""
        (frame_positions, offset_table_type) = self._parse_pixel_data()
        self._offset_table_type = offset_table_type
        return frame_positions

    @property
    def frame_count(self) -> int:
        """Return number of frames"""
        return self.dataset.frame_count

    def get_filepointer(self, frame_index: int) -> Tuple[DicomFileLike, int, int]:
        """Return file pointer, frame position, and frame length for frame
        number.

        Parameters
        ----------
        frame_index: int
            Frame, including concatenation offset, to get.

        Returns
        ----------
        Tuple[WsiDicomFileLike, int, int]:
            File pointer, frame offset and frame length in number of bytes
        """
        frame_index -= self.frame_offset
        frame_position, frame_length = self.frame_positions[frame_index]
        return self._file, frame_position, frame_length

    def _get_offset_table_type(self) -> OffsetTableType:
        """
        Parse file for basic (BOT) or extended offset table (EOT). Return if file has
        EOT and do not check that the BOT is empty (it should be empty according to
        specifications, but we do not care if it not).
        """
        if self.transfer_syntax in UncompressedTransferSyntaxes:
            return OffsetTableType.NONE
        self._file.seek(self._pixel_data_position)
        pixel_data_or_eot_tag = Tag(self._file.read_tag())
        if pixel_data_or_eot_tag == Tag("ExtendedOffsetTable"):
            self._read_tag_vr()
            eot_length = self._read_tag_length()
            self._file.seek(eot_length, 1)
            self._read_eot_lengths_tag()
            return OffsetTableType.EXTENDED
        self._validate_pixel_data_start(pixel_data_or_eot_tag, False)
        bot_length = self._read_bot_length()
        if bot_length is not None:
            return OffsetTableType.BASIC
        return OffsetTableType.EMPTY

    def _read_bot_length(self) -> Optional[int]:
        """Read the length of the basic table offset (BOT). Returns None if BOT
        is empty.

        Returns
        ----------
        Optional[int]
            BOT length.
        """
        BOT_BYTES = 4
        if self._file.read_tag() != ItemTag:
            raise WsiDicomFileError(
                self._file, "Basic offset table did not start with an ItemTag"
            )
        bot_length = self._file.read_UL()
        if bot_length == 0:
            return None
        elif bot_length % BOT_BYTES:
            raise WsiDicomFileError(
                self._file,
                f"Basic offset table should be a multiple of {BOT_BYTES} bytes",
            )
        return bot_length

    def _read_bot(self) -> Optional[bytes]:
        """Read basic table offset (BOT). Returns None if BOT is empty. Filepoiter
        should be positionen to pixel data.

        Returns
        ----------
        Optional[bytes]
            BOT in bytes.
        """
        bot_length = self._read_bot_length()
        if bot_length is None:
            return None
        bot = self._file.read(bot_length)
        return bot

    def _read_eot_length(self) -> int:
        """Read the length of the extended table offset (EOT).

        Returns
        ----------
        int
            EOT length.
        """
        EOT_BYTES = 8
        eot_length = self._read_tag_length()
        if eot_length == 0:
            raise WsiDicomFileError(
                self._file, "Expected Extended offset table present but empty"
            )
        elif eot_length % EOT_BYTES:
            raise WsiDicomFileError(
                self._file,
                "Extended offset table should be a multiple of " f"{EOT_BYTES} bytes",
            )
        return eot_length

    def _read_eot_lengths_tag(self):
        """Skip over the length of the extended table offset lengths tag."""
        eot_lenths_tag = self._file.read_tag()
        if eot_lenths_tag != Tag("ExtendedOffsetTableLengths"):
            raise WsiDicomFileError(
                self._file,
                "Expected Extended offset table lengths tag after reading "
                f"Extended offset table, found {eot_lenths_tag}",
            )
        self._read_tag_vr()
        length = self._read_tag_length()
        # Jump over EOT lengths for now
        self._file.seek(length, 1)

    def _read_eot(self) -> bytes:
        """Read extended table offset (EOT) and EOT lengths. Filepointer should be
        positionend to extended offset table.

        Returns
        ----------
        bytes
            EOT in bytes.
        """
        eot_tag = Tag(self._file.read_tag())
        if eot_tag != Tag("ExtendedOffsetTable"):
            raise ValueError(f"Expected ExtendedOffsetTable tag, got {eot_tag}")
        self._read_tag_vr()
        eot_length = self._read_eot_length()
        # Read the EOT into bytes
        eot = self._file.read(eot_length)
        # Read EOT lengths tag
        self._read_eot_lengths_tag()
        return eot

    def _parse_table(
        self, table: bytes, table_type: OffsetTableType, pixels_start: int
    ) -> List[Tuple[int, int]]:
        """Parse table with offsets (BOT or EOT).

        Parameters
        ----------
        table: bytes
            BOT or EOT as bytes
        table_type: OffsetTableType
            Type of table, 'bot' or 'eot'.
        pixels_start: int
            Position of first frame item in pixel data.

        Returns
        ----------
        List[Tuple[int, int]]
            A list with frame positions and frame lengths.
        """
        if not self._file.is_little_endian:
            raise WsiDicomFileError(
                self._file, "Big endian not supported for BOT or EOT"
            )

        if table_type == OffsetTableType.BASIC:
            bytes_per_item = 4
            mode = "<L"
        elif table_type == OffsetTableType.EXTENDED:
            bytes_per_item = 8
            mode = "<Q"
        else:
            raise ValueError("table type should be 'bot' or 'eot'")
        table_length = len(table)
        TAG_BYTES = 4
        LENGTH_BYTES = 4
        positions: List[Tuple[int, int]] = []
        # Read through table to get offset and length for all but last item
        # All read offsets are for item tag of frame and relative to first
        # frame in pixel data.
        this_offset: int = unpack(mode, table[0:bytes_per_item])[0]
        if this_offset != 0:
            raise ValueError("First item in table should be at offset 0")
        for index in range(bytes_per_item, table_length, bytes_per_item):
            next_offset = unpack(mode, table[index : index + bytes_per_item])[0]
            offset = this_offset + TAG_BYTES + LENGTH_BYTES
            length = next_offset - offset
            if length == 0 or length % 2:
                raise WsiDicomFileError(self._file, "Invalid frame length")
            positions.append((pixels_start + offset, length))
            this_offset = next_offset

        # Go to last frame in pixel data and read the length of the frame
        self._file.seek(pixels_start + this_offset)
        if self._file.read_le_tag() != ItemTag:
            raise WsiDicomFileError(self._file, "Expected ItemTag in PixelData")
        length: int = self._file.read_leUL()
        if length == 0 or length % 2:
            raise WsiDicomFileError(self._file, "Invalid frame length")
        offset = this_offset + TAG_BYTES + LENGTH_BYTES
        positions.append((pixels_start + offset, length))

        return positions

    def _read_positions_from_pixeldata(self) -> List[Tuple[int, int]]:
        """Get frame positions and length from sequence of frames that ends
        with a tag not equal to ItemTag. fp needs to be positioned after the
        BOT.
        Each frame contains:
        item tag (4 bytes)
        item length (4 bytes)
        item data (item length)
        The position of item data and the item length is stored.

        Returns
        ----------
        list[tuple[int, int]]
            A list with frame positions and frame lengths
        """
        TAG_BYTES = 4
        LENGTH_BYTES = 4
        positions: List[Tuple[int, int]] = []
        frame_position = self._file.tell()
        # Read items until sequence delimiter
        while self._file.read_le_tag() == ItemTag:
            # Read item length
            length: int = self._file.read_leUL()
            if length == 0 or length % 2:
                raise WsiDicomFileError(self._file, "Invalid frame length")
            positions.append((frame_position + TAG_BYTES + LENGTH_BYTES, length))
            # Jump to end of frame
            self._file.seek(length, 1)
            frame_position = self._file.tell()
        self._read_sequence_delimiter()
        return positions

    def _create_frame_positions_for_uncapsulated_data(
        self, offset: int
    ) -> List[Tuple[int, int]]:
        """Create frame positions for uncapsulated data.

        Parameters
        ----------
        offset: int
            Offset to first frame in pixel data.

        Returns
        ----------
        List[Tuple[int, int]]
            A list with frame positions and frame lengths.
        """
        frame_size = (
            self.dataset.tile_size.area
            * self.dataset.samples_per_pixel
            * (self.dataset.bits // 8)
        )
        return [
            (offset + index * frame_size, frame_size)
            for index in range(self.frame_count)
        ]

    def _read_sequence_delimiter(self):
        """Check if last read tag was a sequence delimiter.
        Raises WsiDicomFileError otherwise.
        """
        TAG_BYTES = 4
        self._file.seek(-TAG_BYTES, 1)
        if self._file.read_le_tag() != SequenceDelimiterTag:
            raise WsiDicomFileError(self._file, "No sequence delimiter tag")

    def read_frame(self, frame_index: int) -> bytes:
        """Return frame data from pixel data by frame index.

        Parameters
        ----------
        frame_index: int
            Frame, including concatenation offset, to get.

        Returns
        ----------
        bytes
            The frame as bytes
        """
        fp, frame_position, frame_length = self.get_filepointer(frame_index)
        with self._lock:
            fp.seek(frame_position, 0)
            frame: bytes = fp.read(frame_length)
        return frame

    def _validate_pixel_data_start(
        self, tag: Union[BaseTag, Tuple[int, int]], defined_length: bool
    ):
        """Check that pixel data tag is present and that the tag length is
        set as undefined. Raises WsiDicomFileError otherwise.

        Parameters
        ----------
        tag: Union[BaseTag, Tuple[int, int]]
            Tag that should be pixel data tag.
        defined_length: bool
            If length of pixel data should be defined.
        """
        if tag != Tag("PixelData"):
            WsiDicomFileError(self._file, "Expected PixelData tag")
        self._read_tag_vr()
        length = self._read_tag_length()
        if defined_length:
            expected_length = (
                self.dataset.tile_size.area
                * self.dataset.samples_per_pixel
                * (self.dataset.bits // 8)
                * self.frame_count
            )
            if not length == expected_length:
                raise WsiDicomFileError(
                    self._file,
                    f"Expected {expected_length} length when reading Pixel data, got {length}.",
                )
        else:
            if length != 0xFFFFFFFF:
                raise WsiDicomFileError(
                    self._file,
                    f"Expected undefined length when reading Pixel data, got {length}.",
                )

    def _parse_pixel_data(self) -> Tuple[List[Tuple[int, int]], OffsetTableType]:
        """Parse file pixel data, reads frame positions.

        An EOT can be present before the pixel data, and must
        then not be empty. A BOT most always be the first item in the Pixel
        data, but can be empty (zero length). If EOT is used BOT should be empty.

        First search to pixel data position, which is either EOT tag or PixelData tag.
        If EOT read the EOT. For all cases validate that the filepointer now is at the
        PixelData tag. If BOT read the BOT, otherwise skip the BOT. If EOT nor BOT has
        been read, parse frame positions from pixel data. Otherwise parse frame
        positions from EOT or BOT. Finally check that the number of read frames equals
        the specified number of frames, otherwise frames are fragmented which we dont
        support.

        Returns
        ----------
        Tuple[List[Tuple[int, int]], OffsetTableType]
            List of frame positions and lengths, and table type.
        """
        table_type = self.offset_table_type
        table = None
        self._file.seek(self._pixel_data_position)
        if table_type == OffsetTableType.EXTENDED:
            table = self._read_eot()

        self._validate_pixel_data_start(
            Tag(self._file.read_tag()), table_type is OffsetTableType.NONE
        )
        if table_type == OffsetTableType.BASIC:
            table = self._read_bot()
        elif table_type == OffsetTableType.EMPTY:
            self._read_bot_length()

        if table_type == OffsetTableType.NONE:
            frame_positions = self._create_frame_positions_for_uncapsulated_data(
                self._file.tell()
            )
        elif table is None:
            frame_positions = self._read_positions_from_pixeldata()
        else:
            frame_positions = self._parse_table(table, table_type, self._file.tell())

        if len(frame_positions) < self.frame_count:
            raise WsiDicomFileError(
                self._file,
                (
                    f"ImageData contained less frames {len(frame_positions)} than "
                    f"NumberOfFrames {self.frame_count}."
                ),
            )
        if len(frame_positions) > self.frame_count:
            raise WsiDicomFileError(
                self._file,
                (
                    f"ImageData contained more fragments {len(frame_positions)} than "
                    f"NumberOfFrames {self.frame_count} and fragmented frames are not "
                    "supported."
                ),
            )

        return frame_positions, table_type



#    Copyright 2023 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""Module with settings for encoding image data."""

from abc import ABCMeta, abstractmethod
from dataclasses import dataclass
from enum import Enum
from typing import Literal, Union
from pydicom.uid import (
    JPEG2000,
    UID,
    JPEG2000Lossless,
    JPEGBaseline8Bit,
    JPEGExtended12Bit,
    JPEGLosslessP14,
    JPEGLosslessSV1,
    JPEGLSLossless,
    JPEGLSNearLossless,
    ExplicitVRLittleEndian,
    ExplicitVRBigEndian,
    ImplicitVRLittleEndian,
    RLELossless,
)


class Channels(Enum):
    """Color channels in encoded image."""

    GRAYSCALE = "grayscale"
    YBR = "ybr"
    RGB = "rgb"

    @classmethod
    def from_photometric_interpretation(
        cls, photometric_interpretation: str
    ) -> "Channels":
        """Return channels matching photometric interpretation."""
        ybr_photometric_interpretations = [
            "YBR_FULL",
            "YBR_FULL_422",
            "YBR_PARTIAL_422",
            "YBR_ICT",
            "YBR_RCT",
            "YBR",
        ]
        if photometric_interpretation == "MONOCHROME2":
            return cls.GRAYSCALE
        if photometric_interpretation in ybr_photometric_interpretations:
            return cls.YBR
        if photometric_interpretation == "RGB":
            return cls.RGB
        raise ValueError(
            f"Unsupported photometric interpretation: {photometric_interpretation}."
        )


class Subsampling(Enum):
    """JPEG chroma subsampling settings."""

    R444 = "Full horizontal, full vertical (4:4:4)"
    R422 = "Half horisontal, full vertical (4:2:2)"
    R420 = "Half horisontal, half vertical (4:2:0)"
    R411 = "Quarter horisontal, full vertical (4:1:1)"
    R440 = "Full horizontal, half vertical (4:4:0)"


class Settings(metaclass=ABCMeta):
    def __init__(self, bits: int, channels: Channels):
        """
        Initialize encoding settings.

        Parameters:
        ----------
            bits: int
                Number of bits per pixel.
            channels: Channels
                Color channels in encoded image."""
        self._bits = bits
        self._channels = channels

    @property
    def bits(self) -> int:
        """Return bits."""
        return self._bits

    @property
    def channels(self) -> Channels:
        """Return channels."""
        return self._channels

    @property
    @abstractmethod
    def transfer_syntax(self) -> UID:
        """Return transfer syntax."""
        raise NotImplementedError()

    @property
    @abstractmethod
    def extension(self) -> str:
        """Return file extension."""
        raise NotImplementedError()

    @property
    @abstractmethod
    def photometric_interpretation(self) -> str:
        """Return photometric interpretation."""
        raise NotImplementedError()

    @property
    def allocated_bits(self) -> int:
        """Return allocated bits."""
        return self.bits // 8 * 8

    @property
    def high_bit(self) -> int:
        """Return high bit."""
        return self.bits - 1

    @property
    def samples_per_pixel(self) -> int:
        """Return samples per pixel."""
        return 1 if self.channels == Channels.GRAYSCALE else 3

    @classmethod
    def create(
        cls,
        transfer_syntax: UID,
        bits: int,
        photometric_interpretation: str,
    ) -> "Settings":
        """Create settings based on properties dataset and transfer syntax.

        Parameters:
        ----------
            transfer_syntax: UID
                Transfer syntax to create settings for.
            bits: int
                Number of bits per pixel.
            photometric_interpretation: str
                Photometric interpretation of image.

        Returns:
        ----------
            Settings
                Settings matching dataset and transfer syntax.
        """
        jpeg_transfer_syntaxes = [
            JPEGBaseline8Bit,
            JPEGExtended12Bit,
        ]
        jpeg_lossless_transfer_syntaxes = [
            JPEGLosslessP14,
            JPEGLosslessSV1,
        ]
        jpeg_ls_transfer_syntaxes = [
            JPEGLSLossless,
            JPEGLSNearLossless,
        ]
        jpeg_2000_transfer_syntaxes = [
            JPEG2000Lossless,
            JPEG2000,
        ]
        uncompressed_transfer_syntaxes = [
            ImplicitVRLittleEndian,
            ExplicitVRLittleEndian,
            ExplicitVRBigEndian,
        ]
        channels = Channels.from_photometric_interpretation(photometric_interpretation)
        if transfer_syntax in jpeg_transfer_syntaxes:
            return JpegSettings(bits=bits, channels=channels)
        if transfer_syntax in jpeg_lossless_transfer_syntaxes:
            return JpegLosslessSettings(bits=bits, channels=channels)
        if transfer_syntax in jpeg_ls_transfer_syntaxes:
            if channels != Channels.GRAYSCALE:
                raise ValueError("JPEG-LS encoder only supports grayscale images.")
            if transfer_syntax == JPEGLSNearLossless:
                return JpegLsLosslessSettings(level=1, bits=bits)
            return JpegLsLosslessSettings(level=0, bits=bits)
        if transfer_syntax in jpeg_2000_transfer_syntaxes:
            if transfer_syntax == JPEG2000:
                return Jpeg2kSettings(bits=bits, channels=channels)
            return Jpeg2kSettings(level=0, bits=bits, channels=channels)
        if transfer_syntax == RLELossless:
            return RleSettings(bits=bits, channels=channels)
        if transfer_syntax in uncompressed_transfer_syntaxes:
            if channels == Channels.YBR:
                raise ValueError("Numpy encoder does not support ybr images.")
            return NumpySettings(
                bits=bits,
                channels=channels,
                little_endian=transfer_syntax.is_little_endian,
                implicit_vr=transfer_syntax.is_implicit_VR,
            )
        raise ValueError(f"Unsupported transfer syntax: {transfer_syntax}.")


class JpegSettings(Settings):
    """JPEG encoding settings."""

    def __init__(
        self,
        quality: int = 80,
        bits: int = 8,
        channels: Channels = Channels.YBR,
        subsampling: Subsampling = Subsampling.R422,
    ):
        """
        Initialize JPEG encoding settings.

        Parameters:
        ----------
            quality: int = 80
                JPEG quality factor. 0-100, recommended to not use higher than 95.
            bits: int
                Number of bits per pixel.
            channels: Channels = Channels.YBR
                Color channels in encoded image.
            subsampling: Subsampling = Subsampling.R420
                Chroma subsampling settings.
        """
        self._quality = quality
        self._subsampling = subsampling
        super().__init__(bits, channels)

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}(quality={self.quality}, bits={self.bits}, "
            f"channels={self.channels}, subsampling={self.subsampling})"
        )

    @property
    def quality(self) -> int:
        """Return quality."""
        return self._quality

    @property
    def subsampling(self) -> Subsampling:
        """Return subsampling."""
        return self._subsampling

    @property
    def transfer_syntax(self) -> UID:
        if self.bits == 8:
            return JPEGBaseline8Bit
        return JPEGExtended12Bit

    @property
    def photometric_interpretation(self) -> str:
        if self.channels == Channels.GRAYSCALE:
            return "MONOCHROME2"
        if self.channels == Channels.YBR:
            return "YBR_FULL_422"
        return "RGB"

    @property
    def extension(self) -> str:
        return ".jpg"


class JpegLosslessSettings(Settings):
    def __init__(
        self,
        predictor: int = 1,
        bits: int = 8,
        channels: Channels = Channels.YBR,
    ):
        """
        Initialize JPEG lossless encoding settings.

        Parameters:
        ----------
            predictor: int = 1
                JPEG predictor setting. 1-7. Use 1 for SV1.
            bits: int = 8
                Number of bits per pixel.
            channels: Channels = Channels.YBR
                Color channels in encoded image.
        """
        self._predictor = predictor
        super().__init__(bits, channels)

    def __repr__(self):
        return (
            f"{self.__class__.__name__}(predictor={self.predictor}, bits={self.bits}, "
            f"channels={self.channels})"
        )

    @property
    def predictor(self) -> int:
        """Return predictor."""
        return self._predictor

    @property
    def transfer_syntax(self) -> UID:
        if self.predictor is None:
            return JPEGLosslessSV1
        return JPEGLosslessP14

    @property
    def photometric_interpretation(self) -> str:
        if self.channels == Channels.GRAYSCALE:
            return "MONOCHROME2"
        if self.channels == Channels.YBR:
            return "YBR_FULL_422"
        return "RGB"

    @property
    def extension(self) -> str:
        return ".jpg"


@dataclass
class JpegLsLosslessSettings(Settings):
    def __init__(
        self,
        level: int = 0,
        bits: int = 8,
    ):
        """
        Initialize JPEG-LS lossless encoding settings. Only supports grayscale images.

        Parameters:
        ----------
            level: int = 0
                JPEG-LS near-lossless level. 0-9. 0 is lossless.
            bits: int = 8
                Number of bits per pixel.
        """
        self._level = level
        super().__init__(bits, Channels.GRAYSCALE)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(level={self.level}, bits={self.bits})"

    @property
    def level(self) -> int:
        """Return level."""
        return self._level

    @property
    def transfer_syntax(self) -> UID:
        if self.level == 0:
            return JPEGLSLossless
        return JPEGLSNearLossless

    @property
    def photometric_interpretation(self) -> str:
        return "MONOCHROME2"

    @property
    def extension(self) -> str:
        return ".jls"


class Jpeg2kSettings(Settings):
    def __init__(
        self, level: int = 80, bits: int = 8, channels: Channels = Channels.YBR
    ):
        """
        Initialize JPEG 2000 encoding settings.

        Parameters:
        ----------
            level: int = 80
                JPEG 2000 compression level. Set to < 1 or > 1000 for lossless.
            bits: int = 8
                Number of bits per pixel.
            channels: Channels = Channels.YBR
                Color channels in encoded image.
        """
        self._level = level
        super().__init__(bits, channels)

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}(level={self.level}, bits={self.bits}, "
            f"channels={self.channels})"
        )

    @property
    def level(self) -> int:
        """Return level."""
        return self._level

    @property
    def transfer_syntax(self) -> UID:
        if self.level < 1 or self.level > 1000:
            return JPEG2000Lossless
        return JPEG2000

    @property
    def photometric_interpretation(self) -> str:
        if self.channels == Channels.GRAYSCALE:
            return "MONOCHROME2"
        if self.channels == Channels.YBR:
            if self.level is None:
                return "YBR_RCT"
            return "YBR_ICT"
        return "RGB"

    @property
    def extension(self) -> str:
        return ".jp2"


class NumpySettings(Settings):
    def __init__(
        self,
        bits: int = 8,
        channels: Union[
            Literal[Channels.GRAYSCALE], Literal[Channels.RGB]
        ] = Channels.RGB,
        little_endian: bool = True,
        implicit_vr: bool = False,
    ):
        """
        Initialize numpy encoding settings.

        Parameters:
        ----------
            bits: int = 8
                Number of bits per pixel.
            channels: Union[
                Literal[Channels.GRAYSCALE], Literal[Channels.RGB]
            ] = Channels.RGB,
                Color channels in encoded image.
            little_endian: bool = True
                Endianness of encoded image.
            implicit_vr: bool = False
                VR encoding of transfer syntax.
        """
        self._little_endian = little_endian
        self._implicit_vr = implicit_vr
        super().__init__(bits, channels)

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}(bits={self.bits}, channels={self.channels}, "
            f"little_endian={self.little_endian}, implicit_vr={self.implicit_vr})"
        )

    @property
    def little_endian(self) -> bool:
        """Return little endian."""
        return self._little_endian

    @property
    def implicit_vr(self) -> bool:
        """Return implicit vr."""
        return self._implicit_vr

    @property
    def transfer_syntax(self) -> UID:
        if self.little_endian:
            if self.implicit_vr:
                return ImplicitVRLittleEndian
            return ExplicitVRLittleEndian
        return ExplicitVRBigEndian

    @property
    def photometric_interpretation(self) -> str:
        if self.channels == Channels.GRAYSCALE:
            return "MONOCHROME2"
        return "RGB"

    @property
    def extension(self) -> str:
        return ".raw"


class RleSettings(Settings):
    def __init__(self, bits: int = 8, channels: Channels = Channels.RGB):
        super().__init__(bits, channels)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(bits={self.bits}, channels={self.channels})"

    @property
    def transfer_syntax(self) -> UID:
        return RLELossless

    @property
    def photometric_interpretation(self) -> str:
        if self.channels == Channels.GRAYSCALE:
            return "MONOCHROME2"
        if self.channels == Channels.YBR:
            return "YBR_FULL"
        return "RGB"

    @property
    def extension(self) -> str:
        return ".rle"



#    Copyright 2021, 2022, 2023 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from enum import Enum
from pathlib import Path
from typing import BinaryIO, Optional

from pydicom.filebase import DicomFileLike
from pydicom.tag import BaseTag


class OffsetTableType(Enum):
    """Offset table type."""

    NONE = "none"
    EMPTY = "empty"
    BASIC = "BOT"
    EXTENDED = "EOT"

    @classmethod
    def from_string(cls, offset_table: Optional[str]) -> "OffsetTableType":
        """Return OffsetTableType parsed from string."""
        if offset_table is None:
            return OffsetTableType.NONE
        if offset_table.strip().lower() == "empty":
            return OffsetTableType.EMPTY
        if offset_table.strip().lower() == "eot":
            return OffsetTableType.EXTENDED
        if offset_table.strip().lower() == "bot":
            return OffsetTableType.BASIC
        raise ValueError(f"Unknown offset table type: {offset_table}")


class WsiDicomFileBase:
    """Base class for reading or writing DICOM WSI file."""

    def __init__(
        self, stream: BinaryIO, filepath: Optional[Path] = None, owned: bool = False
    ):
        """
        Create a WsiDicomFileBase.

        Parameters
        ----------
        stream: BinaryIO
            Stream to open.
        filepath: Optional[Path] = None
            Optional filepath of stream.
        owned: bool = False
            If the stream should be closed by this instance.
        """
        self._file = DicomFileLike(stream)
        self._filepath = filepath
        self._owned = owned
        self.__enter__()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def __str__(self) -> str:
        return self.pretty_str()

    def pretty_str(self, indent: int = 0, depth: Optional[int] = None) -> str:
        return f"File of stream: {self._file.name}"

    @property
    def filepath(self) -> Optional[Path]:
        """Return filepath."""
        return self._filepath

    def _read_tag_length(self) -> int:
        """Read tag length."""
        return self._file.read_UL()

    def _read_tag_vr(self) -> None:
        """Read tag VR if implicit VRi."""
        if not self._file.is_implicit_VR:
            self._file.read(4, need_exact_length=True)

    def _check_tag_and_length(self, tag: BaseTag, length: int, with_vr: bool) -> None:
        """Check if tag at position is expected tag with expected length.

        Parameters
        ----------
        tag: BaseTag
            Expected tag.
        length: int
            Expected length.
        with_vr: bool
            If tag is expected to have VR.

        """
        read_tag = self._file.read_tag()
        if tag != read_tag:
            raise ValueError(f"Found tag {read_tag} expected {tag}.")
        if with_vr:
            self._read_tag_vr()
        read_length = self._read_tag_length()
        if length != read_length:
            raise ValueError(f"Found length {read_length} expected {length}.")

    def close(self, force: Optional[bool] = False) -> None:
        """Close the file if owned by instance or forced."""
        if self._owned or force:
            self._file.close()



#    Copyright 2021, 2022, 2023 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.


from datetime import datetime
from pathlib import Path
from struct import pack
from typing import (
    BinaryIO,
    Dict,
    Iterable,
    Iterator,
    List,
    Optional,
    Sequence,
    Tuple,
)

from pydicom.dataset import Dataset, FileMetaDataset, validate_file_meta
from pydicom.encaps import itemize_frame
from pydicom.filewriter import write_dataset, write_file_meta_info
from pydicom.tag import ItemTag, SequenceDelimiterTag, Tag
from pydicom.uid import UID, UncompressedTransferSyntaxes

from wsidicom.file.wsidicom_file_base import OffsetTableType, WsiDicomFileBase
from wsidicom.geometry import Point, Region, Size
from wsidicom.instance import ImageData
from wsidicom.instance.dataset import WsiDataset
from wsidicom.thread import ConditionalThreadPoolExecutor
from wsidicom.uid import WSI_SOP_CLASS_UID


class WsiDicomFileWriter(WsiDicomFileBase):
    """Writer for DICOM WSI files."""

    def __init__(
        self,
        stream: BinaryIO,
        little_endian: bool = True,
        implicit_vr: bool = False,
        filepath: Optional[Path] = None,
        owned: bool = False,
    ) -> None:
        """
        Create a writer for DICOM WSI data.

        Parameters
        ----------
        stream: BinaryIO
            Stream to open.
        little_endian: bool = True
            If the file should be written in little endian.
        implicit_vr: bool = False
            If the file should be written with implicit VR.
        filepath: Optional[Path] = None
            Optional filepath of stream.
        owned: bool = False
            If the stream should be closed by this instance.
        """
        super().__init__(stream, filepath, owned)
        self._file.is_little_endian = little_endian
        self._file.is_implicit_VR = implicit_vr

    @classmethod
    def open(cls, file: Path, transfer_syntax: UID) -> "WsiDicomFileWriter":
        """Open file in path as WsiDicomFileWriter.

        Parameters
        ----------
        file: Path
            Path to file.
        transfer_syntax: UID
            Transfer syntax to use for file.

        Returns
        ----------
        WsiDicomFileWriter
            WsiDicomFileWriter for file.
        """
        stream = open(
            file,
            "w+b",
        )
        return cls(
            stream,
            transfer_syntax.is_little_endian,
            transfer_syntax.is_implicit_VR,
            file,
            True,
        )

    def write(
        self,
        uid: UID,
        transfer_syntax: UID,
        dataset: WsiDataset,
        data: Dict[Tuple[str, float], ImageData],
        workers: int,
        chunk_size: int,
        offset_table: OffsetTableType,
        instance_number: int,
        scale: int = 1,
    ) -> None:
        """
        Write data to file.

        Parameters
        ----------
        uid: UID
            Instance UID for file.
        transfer_syntax: UID.
            Transfer syntax for file
        dataset: WsiDataset
            Dataset to write (excluding pixel data).
        data: Dict[Tuple[str, float], ImageData]
            Pixel data to write.
        workers: int
            Number of workers to use for writing pixel data.
        chunk_size: int
            Number of frames to give each worker.
        offset_table: OffsetTableType
            Offset table to use, 'bot' basic offset table, 'eot' extended
            offset table, None - no offset table.
        instance_number: int
            Instance number for file.
        scale: int = 1
            Scale factor.

        """
        if transfer_syntax in UncompressedTransferSyntaxes:
            offset_table = OffsetTableType.NONE
        elif offset_table is OffsetTableType.NONE:
            offset_table = OffsetTableType.EMPTY
        self._write_preamble()
        self._write_file_meta(uid, transfer_syntax)
        dataset.SOPInstanceUID = uid
        dataset.InstanceNumber = instance_number
        self._write_base(dataset)
        if offset_table is OffsetTableType.NONE:
            self._write_unencapsulated_pixel_data(
                dataset, data, workers, chunk_size, scale
            )
        else:
            self._write_encapsulated_pixel_data(
                data, dataset.NumberOfFrames, workers, chunk_size, offset_table, scale
            )

    def _write_encapsulated_pixel_data(
        self,
        data: Dict[Tuple[str, float], ImageData],
        number_of_frames: int,
        workers: int,
        chunk_size: int,
        offset_table: OffsetTableType,
        scale: int = 1,
    ) -> List[int]:
        """Write encapsulated pixel data to file.

        Parameters
        ----------
        data: Dict[Tuple[str, float], ImageData]
            Pixel data to write.
        number_of_frames: int
            Number of frames to write.
        workers: int
            Number of workers to use for writing pixel data.
        chunk_size: int
            Number of frames to give each worker.
        offset_table: OffsetTableType
            Offset table to use.
        scale: int = 1
            Scale factor.

        Returns
        ----------
        List[int]
            List of frame position relative to start of file.
        """
        table_start, pixels_start = self._write_pixel_data_start(
            number_of_frames, offset_table
        )
        frame_positions: List[int] = []
        for (path, z), image_data in sorted(data.items()):
            frame_positions += self._write_pixel_data(
                image_data, True, z, path, workers, chunk_size, scale
            )
        pixels_end = self._file.tell()
        self._write_pixel_data_end()
        if offset_table is not OffsetTableType.EMPTY:
            if table_start is None:
                raise ValueError("Table start should not be None")
            elif offset_table == OffsetTableType.EXTENDED:
                self._write_eot(table_start, pixels_start, frame_positions, pixels_end)
            elif offset_table == OffsetTableType.BASIC:
                self._write_bot(table_start, pixels_start, frame_positions)
        return frame_positions

    def _write_unencapsulated_pixel_data(
        self,
        dataset: WsiDataset,
        data: Dict[Tuple[str, float], ImageData],
        workers: int,
        chunk_size: int,
        scale: int = 1,
    ) -> None:
        """Write unencapsulated pixel data to file.

        Parameters
        ----------
        dataset: WsiDataset
            Dataset with parameters for image to write.
        data: Dict[Tuple[str, float], ImageData]
            Pixel data to write.
        workers: int
            Number of workers to use for writing pixel data.
        chunk_size: int
            Number of frames to give each worker.
        scale: int = 1
            Scale factor.
        """
        length = (
            dataset.tile_size.area
            * dataset.samples_per_pixel
            * (dataset.bits // 8)
            * dataset.frame_count
        )
        self._write_tag("PixelData", "OB", length)
        for (path, z), image_data in sorted(data.items()):
            self._write_pixel_data(
                image_data, False, z, path, workers, chunk_size, scale
            )

    def _write_preamble(self) -> None:
        """Write file preamble to file."""
        preamble = b"\x00" * 128
        self._file.write(preamble)
        self._file.write(b"DICM")

    def _write_file_meta(self, uid: UID, transfer_syntax: UID) -> None:
        """Write file meta dataset to file.

        Parameters
        ----------
        uid: UID
            SOP instance uid to include in file.
        transfer_syntax: UID
            Transfer syntax used in file.
        """
        meta_ds = FileMetaDataset()
        meta_ds.TransferSyntaxUID = transfer_syntax
        meta_ds.MediaStorageSOPInstanceUID = uid
        meta_ds.MediaStorageSOPClassUID = WSI_SOP_CLASS_UID
        validate_file_meta(meta_ds)
        write_file_meta_info(self._file, meta_ds)

    def _write_base(self, dataset: Dataset) -> None:
        """Write base dataset to file.

        Parameters
        ----------
        dataset: Dataset

        """
        now = datetime.now()
        dataset.ContentDate = datetime.date(now).strftime("%Y%m%d")
        dataset.ContentTime = datetime.time(now).strftime("%H%M%S.%f")
        write_dataset(self._file, dataset)

    def _write_tag(
        self, tag: str, value_representation: str, length: Optional[int] = None
    ):
        """Write tag, tag VR and length.

        Parameters
        ----------
        tag: str
            Name of tag to write.
        value_representation: str.
            Value representation (VR) of tag to write.
        length: Optional[int] = None
            Length of data after tag. 'Unspecified' (0xFFFFFFFF) if None.

        """
        if self._file.is_little_endian:
            write_ul = self._file.write_leUL
            write_us = self._file.write_leUS
        else:
            write_ul = self._file.write_beUL
            write_us = self._file.write_beUS
        self._file.write_tag(Tag(tag))
        if not self._file.is_implicit_VR:
            self._file.write(bytes(value_representation, "iso8859"))
            write_us(0)
        if length is not None:
            write_ul(length)
        else:
            write_ul(0xFFFFFFFF)

    def _reserve_eot(self, number_of_frames: int) -> int:
        """Reserve space in file for extended offset table.

        Parameters
        ----------
        number_of_frames: int
            Number of frames to reserve space for.

        """
        table_start = self._file.tell()
        BYTES_PER_ITEM = 8
        eot_length = BYTES_PER_ITEM * number_of_frames
        self._write_tag("ExtendedOffsetTable", "OV", eot_length)
        for _ in range(number_of_frames):
            self._write_unsigned_long_long(0)
        self._write_tag("ExtendedOffsetTableLengths", "OV", eot_length)
        for _ in range(number_of_frames):
            self._write_unsigned_long_long(0)
        return table_start

    def _reserve_bot(self, number_of_frames: int) -> int:
        """Reserve space in file for basic offset table.

        Parameters
        ----------
        number_of_frames: int
            Number of frames to reserve space for.

        """
        table_start = self._file.tell()
        BYTES_PER_ITEM = 4
        tag_lengths = BYTES_PER_ITEM * number_of_frames
        self._file.write_tag(ItemTag)
        self._file.write_leUL(tag_lengths)
        for _ in range(number_of_frames):
            self._file.write_leUL(0)
        return table_start

    def _write_pixel_data_start(
        self, number_of_frames: int, offset_table: OffsetTableType
    ) -> Tuple[Optional[int], int]:
        """Write tags starting pixel data and reserves space for BOT or EOT.

        Parameters
        ----------
        number_of_frames: int
            Number of frames to reserve space for in BOT or EOT.
        offset_table: Optional[str] = 'bot'
            Offset table to use, 'bot' basic offset table, 'eot' extended
            offset table, None - no offset table.

        Returns
        ----------
        Tuple[Optional[int], int]
            Start of table (BOT or EOT) and start of pixel data (after BOT).
        """
        table_start: Optional[int] = None
        if offset_table == OffsetTableType.EXTENDED:
            table_start = self._reserve_eot(number_of_frames)

        # Write pixel data tag
        self._write_tag("PixelData", "OB")

        if offset_table == OffsetTableType.BASIC:
            table_start = self._reserve_bot(number_of_frames)
        elif offset_table == OffsetTableType.EMPTY:
            self._file.write_tag(ItemTag)
            self._file.write_leUL(0)

        pixel_data_start = self._file.tell()

        return table_start, pixel_data_start

    def _write_bot(
        self, bot_start: int, pixel_data_start: int, frame_positions: Sequence[int]
    ) -> None:
        """Write BOT to file.

        Parameters
        ----------
        bot_start: int
            File position of BOT start
        bot_end: int
            File position of BOT end
        frame_positions: Sequence[int]
            List of file positions for frames, relative to file start

        """
        BYTES_PER_ITEM = 4
        # Check that last BOT entry is not over 2^32 - 1
        last_entry = frame_positions[-1] - pixel_data_start
        if last_entry > 2**32 - 1:
            raise NotImplementedError(
                "Image data exceeds 2^32 - 1 bytes "
                "An extended offset table should be used"
            )

        self._file.seek(bot_start)  # Go to first BOT entry
        self._check_tag_and_length(
            ItemTag, BYTES_PER_ITEM * len(frame_positions), False
        )

        for frame_position in frame_positions:  # Write BOT
            self._file.write_leUL(frame_position - pixel_data_start)

    def _write_unsigned_long_long(self, value: int):
        """Write unsigned long long integer (64 bits) as little endian.

        Parameters
        ----------
        value: int
            Value to write.

        """
        self._file.write(pack("<Q", value))

    def _write_eot(
        self,
        eot_start: int,
        pixel_data_start: int,
        frame_positions: Sequence[int],
        last_frame_end: int,
    ) -> None:
        """Write EOT to file.

        Parameters
        ----------
        bot_start: int
            File position of EOT start
        pixel_data_start: int
            File position of EOT end
        frame_positions: Sequence[int]
            List of file positions for frames, relative to file start
        last_frame_end: int
            Position of last frame end.

        """
        BYTES_PER_ITEM = 8
        # Check that last BOT entry is not over 2^64 - 1
        last_entry = frame_positions[-1] - pixel_data_start
        if last_entry > 2**64 - 1:
            raise ValueError(
                "Image data exceeds 2^64 - 1 bytes, likely something is wrong."
            )
        self._file.seek(eot_start)  # Go to EOT table
        self._check_tag_and_length(
            Tag("ExtendedOffsetTable"), BYTES_PER_ITEM * len(frame_positions), True
        )
        for frame_position in frame_positions:  # Write EOT
            relative_position = frame_position - pixel_data_start
            self._write_unsigned_long_long(relative_position)

        # EOT LENGTHS
        self._check_tag_and_length(
            Tag("ExtendedOffsetTableLengths"),
            BYTES_PER_ITEM * len(frame_positions),
            True,
        )
        frame_start = frame_positions[0]
        for frame_end in frame_positions[1:]:  # Write EOT lengths
            frame_length = frame_end - frame_start
            self._write_unsigned_long_long(frame_length)
            frame_start = frame_end

        # Last frame length, end does not include tag and length
        TAG_BYTES = 4
        LENGTH_BYTES = 4
        last_frame_start = frame_start + TAG_BYTES + LENGTH_BYTES
        last_frame_length = last_frame_end - last_frame_start
        self._write_unsigned_long_long(last_frame_length)

    def _write_pixel_data(
        self,
        image_data: ImageData,
        encapsulate: bool,
        z: float,
        path: str,
        workers: int,
        chunk_size: int,
        scale: int = 1,
    ) -> List[int]:
        """Write pixel data to file.

        Parameters
        ----------
        image_data: ImageData
            Image data to read pixel tiles from.
        encapsulate: bool
            If pixel data should be encapsulated.
        z: float
            Focal plane to write.
        path: str
            Optical path to write.
        workers: int
            Maximum number of thread workers to use.
        chunk_size: int
            Chunk size (number of tiles) to process at a time. Actual chunk
            size also depends on minimun_chunk_size from image_data.
        scale: int
            Scale factor (1 = No scaling).

        Returns
        ----------
        List[int]
            List of frame position (position of ItemTag), relative to start of
            file.
        """
        chunked_tile_points = self._chunk_tile_points(image_data, chunk_size, scale)

        if scale == 1:

            def get_tiles_thread(tile_points: Iterable[Point]) -> Iterator[bytes]:
                """Thread function to get tiles as bytes."""
                return image_data.get_encoded_tiles(tile_points, z, path)

            get_tiles = get_tiles_thread
        else:

            def get_scaled_tiles_thread(
                scaled_tile_points: Iterable[Point],
            ) -> Iterator[bytes]:
                """Thread function to get scaled tiles as bytes."""
                return image_data.get_scaled_encoded_tiles(
                    scaled_tile_points, z, path, scale
                )

            get_tiles = get_scaled_tiles_thread

        def write_frame(frame: bytes) -> int:
            """Itemize and write frame to file. Return frame position."""
            position = self._file.tell()
            self._file.write(frame)
            return position

        if encapsulate:

            def encapsulate_tile(tile: bytes) -> Iterable[bytes]:
                return itemize_frame(tile, 1)

        else:

            def encapsulate_tile(tile: bytes) -> Iterable[bytes]:
                return [tile]

        with ConditionalThreadPoolExecutor(max_workers=workers) as pool:
            return [
                write_frame(frame)
                for thread_result in pool.map(get_tiles, chunked_tile_points)
                for tile in thread_result
                for frame in encapsulate_tile(tile)
            ]

    def _chunk_tile_points(
        self, image_data: ImageData, chunk_size: int, scale: int = 1
    ) -> Iterator[Iterator[Point]]:
        """Divide tile positions in image_data into chunks.

        Parameters
        ----------
        image_data: ImageData
            Image data with tiles to chunk.
        chunk_size: int
            Requested chunk size
        scale: int = 1
            Scaling factor (1 = no scaling).

        Returns
        ----------
        Iterator[Iterator[Point]]
            Chunked tile positions
        """
        minimum_chunk_size = getattr(image_data, "suggested_minimum_chunk_size", 1)
        # If chunk_size is less than minimum_chunk_size, use minimum_chunk_size
        # Otherwise, set chunk_size to highest even multiple of
        # minimum_chunk_size
        chunk_size = max(
            minimum_chunk_size, chunk_size // minimum_chunk_size * minimum_chunk_size
        )
        new_tiled_size = image_data.tiled_size.ceil_div(scale)
        # Divide the image tiles up into chunk_size chunks (up to tiled size)
        chunked_tile_points = (
            Region(
                Point(x, y), Size(min(chunk_size, new_tiled_size.width - x), 1)
            ).iterate_all()
            for y in range(new_tiled_size.height)
            for x in range(0, new_tiled_size.width, chunk_size)
        )
        return chunked_tile_points

    def _write_pixel_data_end(self) -> None:
        """Writes tags ending pixel data."""
        self._file.write_tag(SequenceDelimiterTag)
        self._file.write_leUL(0)



#    Copyright 2023 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""Module with codec supporting different formats."""

from typing import Iterable, Optional, Union

import numpy as np
from wsidicom.codec.decoder import Decoder
from wsidicom.codec.encoder import Encoder
from pydicom.uid import (
    UID,
    AllTransferSyntaxes,
)
from PIL.Image import Image as PILImage

from wsidicom.codec.settings import Settings
from wsidicom.geometry import Size


class Codec:
    def __init__(self, decoder: Decoder, encoder: Encoder):
        """Create codec with the given decoder and encoder."""
        self._decoder = decoder
        self._encoder = encoder

    @property
    def encoder(self) -> Encoder:
        return self._encoder

    @property
    def decoder(self) -> Decoder:
        return self._decoder

    @classmethod
    def supported_transfer_syntaxes(
        cls,
        samples_per_pixel: int,
        bits: int,
        photometric_interpretation: str,
        possible_transfer_syntaxes: Optional[Iterable[UID]] = None,
    ) -> Iterable[UID]:
        """Return supported transfer syntaxes for the given parameters.

        Parameters
        ----------
        samples_per_pixel: int
            Samples per pixel of the image.
        bits: int
            Bits per sample of the image.
        photometric_interpretation: str
            Photometric interpretation of the image.
        possible_transfer_syntaxes: Optional[Iterable[UID]]
            Possible transfer syntaxes to check for support.

        Returns
        ----------
        Iterable[UID]
            Supported transfer syntaxes for the given parameters.
        """

        if possible_transfer_syntaxes is None:
            possible_transfer_syntaxes = AllTransferSyntaxes
        return (
            transfer_syntax
            for transfer_syntax in AllTransferSyntaxes
            if cls.is_supported(
                transfer_syntax,
                samples_per_pixel,
                bits,
                photometric_interpretation,
            )
        )

    def decode(self, data: bytes) -> PILImage:
        return self.decoder.decode(data)

    def encode(self, image: Union[PILImage, np.ndarray]) -> bytes:
        return self.encoder.encode(image)

    @classmethod
    def create(
        cls,
        transfer_syntax: UID,
        samples_per_pixel: int,
        bits: int,
        size: Size,
        photometric_interpretation: str,
    ) -> "Codec":
        """Create codec for the given parameters.

        Parameters
        ----------
        transfer_syntax: UID
            Transfer syntax of the image.
        samples_per_pixel: int
            Samples per pixel of the image.
        bits: int
            Bits per sample of the image.
        size: Size
            Size of the image.
        photometric_interpretation: str
            Photometric interpretation of the image.

        Returns
        ----------
        Codec
            Codec for the given parameters.
        """
        decoder = Decoder.create(
            transfer_syntax,
            samples_per_pixel,
            bits,
            size,
            photometric_interpretation,
        )
        encoder = cls._create_encoder(
            transfer_syntax,
            bits,
            photometric_interpretation,
        )
        return cls(decoder, encoder)

    @classmethod
    def is_supported(
        cls,
        transfer_syntax: UID,
        samples_per_pixel: int,
        bits: int,
        photometric_interpretation: str,
    ) -> bool:
        """Return True if codec supports the given parameters.

        Parameters
        ----------
        transfer_syntax: UID
            Transfer syntax of the image.
        samples_per_pixel: int
            Samples per pixel of the image.
        bits: int
            Bits per sample of the image.
        photometric_interpretation: str
            Photometric interpretation of the image.

        Returns
        ----------
        bool
            True if codec supports the given parameters.
        """
        decoder = Decoder.select_decoder(transfer_syntax, samples_per_pixel, bits)
        if decoder is None:
            return False
        try:
            cls._create_encoder(
                transfer_syntax,
                bits,
                photometric_interpretation,
            )
        except Exception:
            return False
        return True

    @classmethod
    def _create_encoder(
        cls,
        transfer_syntax: UID,
        bits: int,
        photometric_interpretation: str,
    ) -> Encoder:
        encoder_settings = Settings.create(
            transfer_syntax,
            bits,
            photometric_interpretation,
        )
        return Encoder.create(encoder_settings)
